% *************************************************************************
% ********************* LearnNSE++ for Regression *************************
% *************************************************************************
% Version 1.0 Nov/08/2014
%
% Developed in MATLAB R2010a
% 
% Author and programmer: Symone Gomes Soares
%
% e-mail: symonesoares@gmail.com or symonesoares@isr.uc.pt
%
% Homepage: http://sites.google.com/site/symonesoares/
%
% Papers: SOARES, Symone Gomes; ARAÚJO, Rui. An On-line Weighted Ensemble 
%         of Regressor Models to Handle Concept Drifts. Engineering  
%         Applications of Artificial Intelligence, v. 37, p. 392-406, 2015. 
%     url: http://www.sciencedirect.com/science/article/pii/S09521976140024
%
%         Elwell, R., Polikar, R., 2011. Incremental Learning of Concept Drift in Nonstationary
%         Environments. IEEE Trans. Neural Netw. 22 (October (10)), 1517–1531.
%
%
% *************************************************************************
% *************************************************************************
% Inputs of the Algorithm:
%
% - input(number of samples x number of input variables) : input matrix
% 
% - output(number of samples x output variable) : output matrix 
%   ps.: it accepts only one output variable
%
% - m       : window's size (m >= 10)
% - Kappa   : discounting factor (0 <= Kappa < 1); Default: 0.2;
% - Theta   : for demarcating correct and incorrect prediction (0 < Theta < 1);
%             Default: 0.05;      
% - Alpha   : factor to include a new model (0 < Alpha < 1);
%             Default: 0.04; 
% - Pruning : ensemble pruning activation; introduce 1 for pruning
% - B       : maximum number of models in the ensemble, if Pruning = 1.
% 
% *************************************************************************
% Outputs of the Algorithm:
%
% CC   : correlation coefficient between predicted outputs and real outputs
% MSE  : mean square errord between predicted outputs and real outputs
% RMSE : root mean square errord between predicted outputs and real outputs
% nM   : number of models of the ensemble
% YR   : real outputs
% YP   : predicted outputs
% loss : loss between predicted outputs and real outputs
% time : running time of the algorithm
%
% *************************************************************************
% *************************************************************************
% *************************************************************************

function [CC, MSE, RMSE, nM, YR, YP, loss, time] = fLearnNSEReg(input, output, m, a, b, Theta, B, Pruning)

    tStart = tic; 
    nM = 1;
    %Build the initial model
    nS                                 = size(input,1);
    Model(nM).X                        = input(1:m,:);
    Model(nM).Y                        = output(1:m,:);
    [~, ~, T_press]                    = KfoldCV(Model(nM).X,Model(nM).Y, 10); % 10 fold crossvaliation
    [Model(nM).minPRESS, Model(nM).lt] = min(T_press);
    [mcxcal,Model(nM).mxcal]           = mncn(Model(nM).X);                  %mean center cal X-block
    [mcycal,Model(nM).mycal]           = mncn(Model(nM).Y);                  %mean center cal Y-block
    Model(nM).bpls                     = simpls(mcxcal,mcycal,Model(nM).lt); %obtain regression vector

    scxtest     = scale(Model(nM).X,Model(nM).mxcal);             %center test X-block
    fk          = rescale(scxtest*Model(nM).bpls(Model(nM).lt,:)',Model(nM).mycal);  %project and rescale predictions
    AREk        = abs((fk' - Model(nM).Y') ./ Model(nM).Y'); 
    IX          = AREk > Theta;
    DEns        = ones(1,m) ./ m; %normalize
    Model(nM).e = sum(DEns(1,IX)) + 0.00001;
    if (Model(nM).e > 0.5)
      Model(nM).b = 1;
      else  Model(nM).b = (Model(nM).e + 0.01) / (1 - Model(nM).e + 0.01);
    end
    c           =  -a * (nM - nM - b);
    Model(nM).w = 1 / (1 + exp(c));
    Model(nM).B = Model(nM).e * Model(nM).b;

    nM     = nM + 1;
    BffIn  = [];
    BffOut = [];
    YP     = [];
    YR     = [];

    for i = (m + 1):nS

        %Receiving a new sample
        xi = input(i,:);
        yi = output(i,:);
        
        % Making predictions;
        [yiEns, ~] = EnsemblePLSAdaBoostLog(Model, xi);
        %fprintf('\nSample [%d], real output: %f, predicted output: %f \n', i, yi, yiEns);
        YP = [YP yiEns];
        YR = [YR yi];
        
        %Creating the Buffer
        BffIn  = [BffIn;  xi];
        BffOut = [BffOut; yi];
        
        %If the Buffer is full...Train a Model...Reset the Buffer
        if (size(BffOut,1) == m)
          %Obtain the predictions of the ensemble over the new data set; 
          fE = zeros(1,m);
          for j = 1:m
             [fE(1,j), ~] = EnsemblePLSAdaBoostLog(Model, BffIn(j,:));
          end
          
          %Compute the error of the existing ensemble on new data
          AREt = abs((fE - BffOut') ./ BffOut');
          
          %Obtain the new distribution...
          DEns  = ones(1,m) ./ m; %normalize
          Error = size(find(AREt > Theta),2);
          if (Error ~= 0)
              Upfactor   = m / Error;
              Downfactor = 1 / Upfactor;
              for j = 1:m
                  if AREt(1,j) > Theta
                      DEns(1,j) = DEns(1,j) * Upfactor;
                  else
                      DEns(1,j) = DEns(1,j) * Downfactor;
                  end
              end
              DEns = DEns ./ sum(DEns);
          end
          
          %Generating a new model....
          Model(nM).X              = BffIn;
          Model(nM).Y              = BffOut;
          [~, ~, T_press]          = KfoldCV(Model(nM).X,Model(nM).Y, 10); % 10 fold crossvaliation
          [Model(nM).minPRESS, Model(nM).lt] = min(T_press);
          [mcxcal,Model(nM).mxcal] = mncn(Model(nM).X);                  %mean center cal X-block
          [mcycal,Model(nM).mycal] = mncn(Model(nM).Y);                  %mean center cal Y-block
          Model(nM).bpls           = simpls(mcxcal,mcycal,Model(nM).lt); %obtain regression vector
          
          %Evaluate all existing models on the new distribution DEns
          for j = 1:size(Model,2)
             scxtest = scale(BffIn,Model(j).mxcal);             %center test X-block
             fk      = rescale(scxtest*Model(j).bpls(Model(j).lt,:)',Model(j).mycal);  %project and rescale predictions
             AREk    = abs((fk' - BffOut') ./ BffOut'); 
             IX      = AREk > Theta;
             Model(j).e = sum(DEns(1,IX)) + 0.00001;
             if j < nM
                if (Model(j).e > 0.5)
                   Model(j).b = [Model(j).b; 1];
                else
                   er         = (Model(j).e + 0.01) / (1 - Model(j).e + 0.01);
                   Model(j).b = [Model(j).b; er];
                end
             else
                if (Model(j).e > 0.5)
                   Model(j).b = 1;
                else
                   Model(j).b = (Model(j).e + 0.01) / (1 - Model(j).e + 0.01);
                end
             end
          end 
          %weighted average...
          for j = 1:(nM-1) 
            c          = -a * (nM - j - b);
            wkt        = 1 / (1 + exp(c));
            Model(j).w = [Model(j).w; wkt];
            sum1       = sum(Model(j).w);
            Model(j).w = Model(j).w / sum1;
          end
          c           =  -a * (nM - nM - b);
          Model(nM).w = 1 / (1 + exp(c));
          for j = 1:nM
             sum1 = 0;
             for y = 1:size(Model(j).w)
                sum1 = Model(j).w(y) *  Model(j).b(y) + sum1;
             end
             Model(j).B = sum1;
          end
          
          if size(Model,2) > B && Pruning == 1
              maxError = -1;
              for j = 1:nM
                  if(Model(j).e >= maxError)
                     maxError = Model(j).e;
                     ix       = j;
                  end
              end
              Model(ix) = [];
              nM        = nM - 1;
          end
          
          BffIn  = [];
          BffOut = [];  
          nM     = nM + 1;
        end
        
    end
    
    time = toc(tStart);
    CC   = corr2(YR,YP);
    MSE  = mse(YP - YR);
    RMSE = mse(YP - YR)^1/2;
    nM   = size(Model,2);
    loss = (YP - YR).^2;
    fprintf('\n\n\nCorrelation coefficient: %f \n', CC);
    fprintf('Mean Squared Error:      %f \n', MSE);
    fprintf('Root Mean Squared Error: %f \n', RMSE);
    fprintf('Total number of models:  %d \n', nM);

end
